import { LobbyAPI, LobbyMessageType, LobbyUserObj, UserModel } from 'wetube-common';
import {
    WSController, WSClientConnection, WSEndpointMiddleware,
    WSEndpointHandler, WSEndpointBindings
} from "../websocket";
import { handshakeMiddleware, getAuthMiddleware } from '../middleware';
import { LobbyService, UserService } from "../services";
import { AuthData, ErrorResponse } from '../contract';

export class LobbyController extends WSController<LobbyAPI> {
    // lobbyID -> userID -> connection
    private lobbyConnections = new Map<string, Map<string, WSClientConnection>>();
    private lobbyMiddleware: WSEndpointMiddleware[] = [];
    private lobbyService: LobbyService;
    private userService: UserService;

    constructor(
        userService: UserService,
        lobbyService: LobbyService,
    ) {
        super();
        this.lobbyService = lobbyService;
        this.userService = userService;

        this.lobbyMiddleware = [
            handshakeMiddleware,
            getAuthMiddleware(this.userService)
        ];
    }

    public register(): WSEndpointBindings<LobbyAPI> {
        return {
            [LobbyMessageType.CREATE]: {
                handler: this.createLobby,
                middleware: [handshakeMiddleware]
            },
            [LobbyMessageType.JOIN]: {
                handler: this.joinLobby,
                middleware: [handshakeMiddleware]
            },
            [LobbyMessageType.LEAVE]: {
                handler: this.leaveLobby,
                middleware: this.lobbyMiddleware
            },
            [LobbyMessageType.DATA_OFFER]: {
                handler: this.dataOffer,
                middleware: this.lobbyMiddleware
            },
            [LobbyMessageType.DATA_ANSWER]: {
                handler: this.dataAnswer,
                middleware: this.lobbyMiddleware
            },
            [LobbyMessageType.ICE_CANDIDATE]: {
                handler: this.iceCandidate,
                middleware: this.lobbyMiddleware
            },
        };
    }

    public createLobby: WSEndpointHandler<LobbyAPI[LobbyMessageType.CREATE]> = async () => {
        const lobby = this.lobbyService.createLobby();
        return {
            lobbyId: lobby.id,
        };
    };

    public joinLobby: WSEndpointHandler<LobbyAPI[LobbyMessageType.JOIN], Partial<AuthData>> = async (req, connection) => {
        const { data, auth } = req.body;

        let user: UserModel | null = null;
        if (auth) {
            // Check if we can get existing user via auth data in request
            try {
                const decoded = this.userService.validateUser(auth);
                user = this.userService.getUser(decoded.id);
            } catch (_) {
                return ErrorResponse.BadRequest;
            }
        }
        if (!user) {
            // If failed - create new user and set token for connection
            user = this.userService.createUser();
            connection.authToken = user.token;
        } else {
            // If user found - leave all current lobbies
            const lobbiesToLeave = this.lobbyService.leaveAllLobbies(user.id);
            lobbiesToLeave.forEach(lobby => {
                this.broadcastToLobby(LobbyMessageType.LEAVE, lobby.id, {
                    lobbyId: lobby.id,
                    users: this.usersToUsersResponseObj(lobby.users)
                });
            });
        }

        const lobby = this.lobbyService.joinLobby(data.lobbyId, user.id);
        if (!lobby) {
            return ErrorResponse.BadRequest;
        }

        const lobbyUsers = this.usersToUsersResponseObj(lobby.users);

        this.addConnection(connection, lobby.id, user.id);
        this.broadcastToLobby(LobbyMessageType.JOIN, lobby.id, {
            lobbyId: lobby.id,
            users: lobbyUsers
        });

        return {
            lobbyId: lobby.id,
            user: {
                id: user.id,
                name: user.name,
                token: user.token
            },
            users: lobbyUsers,
        };
    };

    public leaveLobby: WSEndpointHandler<LobbyAPI[LobbyMessageType.LEAVE], AuthData> = async (req) => {
        const { data } = req.body;
        const userId = req.user.id;

        const lobby = this.lobbyService.leaveLobby(data.lobbyId, userId);
        if (!lobby) {
            return ErrorResponse.BadRequest;
        }

        this.removeConnection(lobby.id, userId);
        this.broadcastToLobby(LobbyMessageType.LEAVE, lobby.id, {
            lobbyId: lobby.id,
            users: this.usersToUsersResponseObj(lobby.users)
        });

        return this.noAck;
    };

    public iceCandidate: WSEndpointHandler<LobbyAPI[LobbyMessageType.ICE_CANDIDATE], AuthData> = async (req) => {
        const { data } = req.body;
        const userId = req.user.id;

        const recipient = this.getConnection(data.lobbyId, data.to);
        if (!recipient) {
            return this.noAck;
        }

        this.sendTo(recipient, LobbyMessageType.ICE_CANDIDATE, {
            lobbyId: data.lobbyId,
            from: userId,
            to: data.to,
            candidate: data.candidate
        });

        return this.noAck;
    };

    public dataOffer: WSEndpointHandler<LobbyAPI[LobbyMessageType.DATA_OFFER], AuthData> = async (req) => {
        const { data } = req.body;
        const userId = req.user.id;

        const recipient = this.getConnection(data.lobbyId, data.to);
        if (!recipient) {
            return this.noAck;
        }

        this.sendTo(recipient, LobbyMessageType.DATA_OFFER, {
            lobbyId: data.lobbyId,
            from: userId,
            to: data.to,
            description: data.description
        });

        return this.noAck;
    };

    public dataAnswer: WSEndpointHandler<LobbyAPI[LobbyMessageType.DATA_ANSWER], AuthData> = async (req) => {
        const { data } = req.body;
        const userId = req.user.id;

        const recipient = this.getConnection(data.lobbyId, data.to);
        if (!recipient) {
            return this.noAck;
        }

        this.sendTo(recipient, LobbyMessageType.DATA_ANSWER, {
            lobbyId: data.lobbyId,
            from: userId,
            to: data.to,
            description: data.description
        });

        return this.noAck;
    };

    public override onDisconnect = (connection: WSClientConnection) => {
        const connData = this.findConnectionData(connection);
        if (!connData) {
            return;
        }
        this.removeConnection(connData.lobbyId, connData.userId);

        const lobby = this.lobbyService.leaveLobby(connData.lobbyId, connData.userId);
        if (!lobby) {
            return;
        }
        this.broadcastToLobby(LobbyMessageType.LEAVE, lobby.id, {
            lobbyId: lobby.id,
            users: this.usersToUsersResponseObj(lobby.users)
        });
    };

    private usersToUsersResponseObj(users: string[]): LobbyUserObj[] {
        return users
            .map(userId => this.userService.getUser(userId))
            .filter((user): user is UserModel => user != null)
            .map(user => ({ id: user.id, name: user.name }));
    }

    private broadcastToLobby<K extends keyof LobbyAPI>(type: K, lobbyId: string, data: LobbyAPI[K]['response']) {
        const lobbyGroup = this.lobbyConnections.get(lobbyId);
        if (!lobbyGroup) {
            return;
        }

        lobbyGroup.forEach(connection => {
            this.sendTo(connection, type, data);
        });
    }

    private getConnection(lobbyId: string, userId: string) {
        return this.lobbyConnections.get(lobbyId)?.get(userId) ?? null;
    }

    private addConnection(connection: WSClientConnection, lobbyId: string, userId: string) {
        if (!this.lobbyConnections.has(lobbyId)) {
            const users = new Map();
            users.set(userId, connection);
            this.lobbyConnections.set(lobbyId, users);
        }
        const lobbyUsers = this.lobbyConnections.get(lobbyId);
        lobbyUsers?.set(userId, connection);
    }

    private removeConnection(lobbyId: string, userId: string) {
        const lobbyGroup = this.lobbyConnections.get(lobbyId);
        lobbyGroup?.delete(userId);
    }

    private findConnectionData(connection: WSClientConnection) {
        let targetLobbyId: string | undefined;
        let targetUserId: string | undefined;

        this.lobbyConnections.forEach((lobbyUsers, lobbyId) => {
            lobbyUsers.forEach((conn, userId) => {
                if (conn.id == connection.id) {
                    targetLobbyId = lobbyId;
                    targetUserId = userId;
                    return;
                }
            });
        });

        return targetLobbyId && targetUserId ? {
            lobbyId: targetLobbyId,
            userId: targetUserId
        } : null;
    }
}