import express from 'express';
import http from 'http';
import { Server as WebSocketServer } from 'ws';

import { WSRouter } from './websocket';
import { UserInMemoryRepository, LobbyInMemoryRepository } from './repositories';
import { UserService, LobbyService } from './services';
import { LobbyController, UserController } from './controllers';
import { AuthDataPayload } from './contract';
import { JWTAuth } from './utils/token';
import { RandomNameGenerator } from './utils/random/RandomNameGenerator';

const app = express();
const server = http.createServer(app);
const wss = new WebSocketServer({ server: server });

// Define repositories
const userRepo = new UserInMemoryRepository();
const lobbyRepo = new LobbyInMemoryRepository();

// Define utils
const jwtAuth = new JWTAuth<AuthDataPayload>();
const randomNameGenerator = new RandomNameGenerator();

// Define services
const userService = new UserService(jwtAuth, userRepo, randomNameGenerator);
const lobbyService = new LobbyService(userRepo, lobbyRepo);

// Define controllers
const lobbyController = new LobbyController(userService, lobbyService);
const userController = new UserController(userService);

// Define router and register
const router = new WSRouter(wss);
router.register(lobbyController);
router.register(userController);

router.start();
server.listen(process.env.PORT || 3000, () => {
    console.log('Server started');
});