import React from 'react';
import { styled } from '@mui/material/styles';
import Slider from '@mui/material/Slider';
import IconButton from '@mui/material/IconButton';
import Icon from '@mui/material/Icon';
import Box from '@mui/material/Box';
import Typography from '@mui/material/Typography';

const ControlsContainer = styled(Box)(({ theme }) => ({
    display: 'flex',
    flexDirection: 'column',
    backgroundColor: `${theme.palette.background.paper}`,
}));

const StyledPlaybackSlider = styled(Slider)(() => ({
    height: 14,
    '& .MuiSlider-thumb': {
        height: 20,
        width: 20,
        backgroundColor: '#fff',
        border: '2px solid currentColor',
        '&:focus, &:hover, &.Mui-active, &.Mui-focusVisible': {
            boxShadow: 'inherit',
        },
        '&:before': {
            display: 'none',
        },
    },
}));

const SecondaryControlsContainer = styled(Box)(() => ({
    display: 'flex',
    alignItems: 'center',
    justifyContent: 'space-between'
}));

const InternalControlsContainer = styled(Box)(() => ({
    display: 'flex',
    alignItems: 'center',
}));

const StyledVolumeSlider = styled(Slider)(() => ({
    width: 60,
    marginLeft: 10,
    '& .MuiSlider-thumb': {
        height: 12,
        width: 12,
        backgroundColor: '#fff',
        border: '2px solid currentColor',
        '&:focus, &:hover, &.Mui-active, &.Mui-focusVisible': {
            boxShadow: 'inherit',
        },
        '&:before': {
            display: 'none',
        },
    },
}));

interface PlayerControlsProps {
    duration: number,
    elapsed: number,
    playing: boolean,
    volume: number,
    onPlay: () => void,
    onPause: () => void,
    onSeek: (value: number | number[]) => void,
    onVolumeChange: (value: number | number[]) => void,
}
export const PlayerControls = (props: PlayerControlsProps) => {
    const {
        duration, elapsed, playing, volume,
        onPlay, onPause, onSeek, onVolumeChange,
    } = props;
    const playbackSliderValue = duration > 0 ? (elapsed / duration) * 100 : 0;

    const onPlaybackSliderChange = (_: Event, value: number | number[]) => onSeek(value);
    const onVolumeSliderChange = (_: Event, value: number | number[]) => onVolumeChange(value);

    const leadingZeroFormat = (value: number) => {
        return value < 10 ? `0${value}` : `${value}`;
    };
    const secondsToTimeFormat = (value: number) => {
        // if value is less than 1h
        if (value < 3600) {
            const minutes = Math.floor(value / 60);
            const seconds = Math.floor(value - (60 * minutes));
            return `${leadingZeroFormat(minutes)}:${leadingZeroFormat(seconds)}`;
        }

        const hours = Math.floor(value / 3600);
        const minutes = Math.floor((value - (3600 * hours)) / 60);
        const seconds = Math.floor(value - (3600 * hours) - (60 * minutes));
        return `${leadingZeroFormat(hours)}:${leadingZeroFormat(minutes)}:${leadingZeroFormat(seconds)}`;
    };

    return (
        <ControlsContainer>
            <StyledPlaybackSlider
                defaultValue={0}
                value={playbackSliderValue}
                onChange={onPlaybackSliderChange}
                aria-label="Default"
                valueLabelDisplay="off"
                size='small'
            />
            <SecondaryControlsContainer>
                <InternalControlsContainer>
                    {
                        playing ?
                            <IconButton onClick={onPause}>
                                <Icon>pause</Icon>
                            </IconButton>
                            :
                            <IconButton onClick={onPlay}>
                                <Icon>play_arrow</Icon>
                            </IconButton>
                    }
                    <IconButton onClick={onPause}>
                        <Icon>volume_up</Icon>
                    </IconButton>
                    <StyledVolumeSlider
                        defaultValue={0}
                        value={volume}
                        onChange={onVolumeSliderChange}
                        aria-label="Default"
                        valueLabelDisplay="auto"
                        size='small'
                    />
                </InternalControlsContainer>
                <Typography
                    noWrap
                    variant="body1"
                    component="div"
                >
                    {`${secondsToTimeFormat(elapsed)} / ${secondsToTimeFormat(duration)}`}
                </Typography>
            </SecondaryControlsContainer>
        </ControlsContainer>
    );
};