import { useEffect, useState, useCallback, useMemo, useRef } from 'react';
import { useSearch, useNavigate, MakeGenerics } from '@tanstack/react-location';
import { LobbyUserObj } from 'wetube-common';
import { AppApiSubscriptions } from '../../../api';
import { useAppApi } from '../../../context';
import { useMediaPlaylist } from './useMediaPlaylist';
import { usePlayer } from './usePlayer';

type MyLocationGenerics = MakeGenerics<{
    Search: {
        lobby: string
    }
}>;

const initialVideoUrl = 'https://www.youtube.com/watch?v=NcRifDitRnU';
export const useJoinLobby = () => {
    // services and utils
    const api = useAppApi();
    const search = useSearch<MyLocationGenerics>();
    const navigate = useNavigate();
    const playlist = useMediaPlaylist();
    const player = usePlayer();

    // state and search
    const [lobbyUsers, setLobbyUsers] = useState<LobbyUserObj[]>([]);
    const [isInitialJoin, setIsInitialJoin] = useState<boolean | undefined>(undefined);
    const prevLobbyUsers = useRef<LobbyUserObj[]>([]);
    const lobbyId = search.lobby;

    useEffect(() => {
        // can't continue to join without lobby id
        if (!lobbyId) {
            return;
        }

        api.getMediaMetadata(initialVideoUrl)
            .then(metadata => {
                if (!metadata) {
                    return;
                }
                playlist.set([metadata]);
            });

        const onJoin = ({ data }: AppApiSubscriptions['join-lobby']) => {
            if (!data) {
                return;
            }

            setIsInitialJoin(prev => prev ? false : prev);
            setLobbyUsers((prev) => {
                prevLobbyUsers.current = prev;
                return data.users;
            });
        };
        const onLeave = ({ data }: AppApiSubscriptions['leave-lobby']) => {
            if (!data) {
                return;
            }

            setLobbyUsers(data.users);
        };
        const onSync = (data: AppApiSubscriptions['sync-playlist-media']) => {
            playlist.set(data.playlist);
        };

        api.subscription
            .subscribe('join-lobby', onJoin)
            .subscribe('leave-lobby', onLeave)
            .subscribe('sync-playlist-media', onSync);

        api.joinLobby(lobbyId)
            .then(res => {
                const { data } = res;
                if (!data || !data.user) {
                    return;
                }

                setIsInitialJoin(() => true);
                setLobbyUsers(data.users);
            });

        return () => {
            api.subscription
                .unsubscribe('join-lobby', onJoin)
                .unsubscribe('leave-lobby', onLeave)
                .unsubscribe('sync-playlist-media', onSync);
        };
    }, []);

    const syncMediaPlaylist = useCallback((id: string) => {
        if (isInitialJoin || isInitialJoin == undefined) {
            return;
        }

        api.syncMediaPlaylist(playlist.items, id);
    }, [isInitialJoin, playlist.items]);

    useEffect(() => {
        const newJoined = lobbyUsers.filter(q => !prevLobbyUsers.current.some(u => q.id == u.id));
        newJoined.forEach(q => {
            syncMediaPlaylist(q.id);
        });
    }, [lobbyUsers, prevLobbyUsers]);

    const leaveLobby = useCallback(() => {
        if (lobbyId) {
            api.leaveLobby(lobbyId);
        }
        navigate({ to: '/', replace: true });
    }, []);

    const currentVideo = useMemo(() => {
        return playlist.items.length > 0 ? playlist.items[0].id : null;
    }, [playlist.items]);

    return {
        playlist,
        player,
        currentVideo,
        lobbyUsers,
        leaveLobby,
    };
};