import { LobbyAPI, LobbyMessageType } from 'wetube-common';
import { ILobbyService, LobbySubscriptions } from './ILobbyService';
import { ISubscription, Subscription, WSClientConnection } from "../../utils";
import { IAuthDataServices } from '../auth-data';

export class LobbyService implements ILobbyService {
    private _subscription: Subscription<LobbySubscriptions>;
    private socket: WSClientConnection<LobbyAPI>;
    private authDataService: IAuthDataServices;

    constructor(socket: WSClientConnection<LobbyAPI>, authDataService: IAuthDataServices) {
        this._subscription = new Subscription();
        this.socket = socket;
        this.authDataService = authDataService;

        this.socket.subscription
            .subscribe(LobbyMessageType.JOIN, this.onJoin)
            .subscribe(LobbyMessageType.LEAVE, this.onLeave)
            .subscribe(LobbyMessageType.DATA_OFFER, this.onDataOffer)
            .subscribe(LobbyMessageType.DATA_ANSWER, this.onDataAnswer)
            .subscribe(LobbyMessageType.ICE_CANDIDATE, this.onIceCandidate);
    }

    public get subscription(): ISubscription<LobbySubscriptions> {
        return this._subscription;
    }

    public create() {
        return this.socket.sendAck()(LobbyMessageType.CREATE, {});
    }

    public join(lobbyId: string) {
        return this.socket.sendAck()(LobbyMessageType.JOIN, {
            lobbyId,
        }).then(res => {
            if (res.data?.user) {
                this.authDataService.saveData(res.data?.user);
            }

            return res;
        });
    }

    public leave(lobbyId: string) {
        this.socket.send()(LobbyMessageType.LEAVE, {
            lobbyId,
        }).then(res => {
            this.authDataService.clearData();
            return res;
        });
    }

    public sendDataOffer(toId: string, lobbyId: string, offer: RTCSessionDescriptionInit) {
        this.socket.send()(LobbyMessageType.DATA_OFFER, {
            lobbyId,
            to: toId,
            description: offer,
        });
    }

    public sendDataAnswer(toId: string, lobbyId: string, answer: RTCSessionDescriptionInit) {
        this.socket.send()(LobbyMessageType.DATA_ANSWER, {
            lobbyId,
            to: toId,
            description: answer,
        });
    }

    public sendIceCandidate(toId: string, lobbyId: string, candidate: RTCIceCandidate) {
        this.socket.send()(LobbyMessageType.ICE_CANDIDATE, {
            lobbyId,
            to: toId,
            candidate,
        });
    }

    private onJoin = (res: LobbySubscriptions['join']) => {
        this._subscription.publish('join', res);
    };
    private onLeave = (res: LobbySubscriptions['leave']) => {
        this._subscription.publish('leave', res);
    };
    private onDataOffer = (res: LobbySubscriptions['data-offer']) => {
        this._subscription.publish('data-offer', res);
    };
    private onDataAnswer = (res: LobbySubscriptions['data-answer']) => {
        this._subscription.publish('data-answer', res);
    };
    private onIceCandidate = (res: LobbySubscriptions['ice-candidate']) => {
        this._subscription.publish('ice-candidate', res);
    };
}