import { IMediaMetadataService, MediaMetadata } from './IMediaMetadataService';

interface IMediaMetadataProvider {
    test(query: string): boolean;
    fetch(query: string): Promise<MediaMetadata | null>
}

interface YoutubeVideoMetadata {
    title: string,
    author_name: string,
    type: string,
    provider_name: string,
    provider_url: string,
    thumbnail_url: string,
}
class YoutubeVideoMetadataProvider implements IMediaMetadataProvider {
    private static URL_REGEX = /http(?:s?):\/\/(?:www\.)?youtu(?:be\.com\/watch\?v=|\.be\/)([\w\-_]*)(&(amp;)?[\w?=]*)?/;

    test(query: string): boolean {
        return YoutubeVideoMetadataProvider.URL_REGEX.test(query);
    }

    async fetch(query: string): Promise<MediaMetadata | null> {
        const url = `https://www.youtube.com/oembed?url=${query}`;

        const id = this.getIdFromQuery(query);
        if (!id) {
            return null;
        }

        const result = await fetch(url);
        if (!result.ok) {
            return null;
        }

        const data = await result.json();
        if (!this.isValidMetadata(data)) {
            return null;
        }

        return {
            id: id,
            title: data.title,
            author: data.author_name,
            thumbnailUrl: data.thumbnail_url,
        };
    }

    private isValidMetadata(data: unknown): data is YoutubeVideoMetadata {
        return typeof data == 'object' && data != null &&
            'title' in data && 'author_name' in data &&
            'type' in data && 'provider_name' in data &&
            'provider_url' in data && 'thumbnail_url' in data;
    }

    private getIdFromQuery(query: string): string | null {
        const results = query.match(YoutubeVideoMetadataProvider.URL_REGEX);
        if (!results || results.length < 2) {
            return null;
        }

        return results[1];
    }
}

export class MediaMetadataService implements IMediaMetadataService {
    private providers: IMediaMetadataProvider[];

    constructor() {
        this.providers = [
            new YoutubeVideoMetadataProvider(),
        ];
    }

    getMetadataByQuery(query: string): Promise<MediaMetadata | null> {
        const provider = this.providers.find(provider => provider.test(query));
        if (!provider) {
            return Promise.resolve(null);
        }

        return provider.fetch(query);
    }

}