import { EmptyObj, AnyObj, WSEndpoint, WSAnyEndpoints, WSPacketBody, WSPacketError } from 'wetube-common';
import { ISubscription, Subscription } from './Subscription';

type ReservedSocketSubscriptions = {
    'open': EmptyObj,
}
type ReservedEndpoints = {
    '.handshake': WSEndpoint<EmptyObj, { id: string }>
}

type WSEndpointRes<T extends AnyObj> = {
    data?: T,
    error?: WSPacketError
};
export type WSRes<T extends WSAnyEndpoints, K extends keyof T> = WSEndpointRes<T[K]['response']>

type ExtractResponses<T extends WSAnyEndpoints> = {
    [K in keyof T]: WSRes<T, K>
}

type WSClientConnectionOptions<Endpoints extends WSAnyEndpoints> = {
    getAuthToken: () => string | null,
    authEndpoints: (keyof Endpoints)[]
};
export class WSClientConnection<Endpoints extends WSAnyEndpoints> {
    private _subscription: Subscription<ReservedSocketSubscriptions & ExtractResponses<Endpoints>>;
    private pendingHandshake: {
        promise: Promise<void>,
        resolve?: () => void,
        reject?: () => void
    };
    private pendingAcks = new Map<number, (v: WSEndpointRes<AnyObj>) => void>();
    private getAuthToken: () => string | null;
    private authEndpoints: (keyof Endpoints)[] = [];
    private id: string | null = null;
    private ws: WebSocket | null = null;
    private isConnected = false;
    private url: string;
    private ackId = 1;


    constructor(url: string, options: WSClientConnectionOptions<Endpoints>) {
        this.url = url;
        this.getAuthToken = options.getAuthToken;
        this.authEndpoints = options.authEndpoints;
        this._subscription = new Subscription();

        this.pendingHandshake = { promise: Promise.resolve() };
        this.pendingHandshake.promise = new Promise<void>((res, rej) => {
            this.pendingHandshake.resolve = res;
            this.pendingHandshake.reject = rej;
        });
    }

    public get subscription(): ISubscription<ReservedSocketSubscriptions & ExtractResponses<Endpoints>> {
        return this._subscription;
    }

    public get connected(): boolean {
        return this.isConnected;
    }

    public connect() {
        this.ws = new WebSocket(this.url);
        this.ws.addEventListener('open', this.onOpen);
        this.ws.addEventListener('close', this.onClose);
        this.ws.addEventListener('error', this.onError);
        this.ws.addEventListener('message', this.onMessage);
    }

    public disconect() {
        if (!this.ws) {
            return;
        }

        this.ws.close();
        this.ws.removeEventListener('open', this.onOpen);
        this.ws.removeEventListener('close', this.onClose);
        this.ws.removeEventListener('error', this.onError);
        this.ws.removeEventListener('message', this.onMessage);
    }

    public send<T extends Endpoints>(): <K extends keyof T>(type: K, data: T[K]['request']) => Promise<WSRes<T, K>> {
        return (type, data) => this.pendingHandshake.promise.then(() => this.internalSend(type, data, false));
    }

    public sendAck<T extends Endpoints>(): <K extends keyof T>(type: K, data: T[K]['request']) => Promise<WSRes<T, K>> {
        return (type, data) => this.pendingHandshake.promise.then(() => this.internalSend(type, data, true));
    }

    private internalSend = <T extends WSAnyEndpoints, K extends keyof T>(type: K, data: T[K]['request'], ack: boolean): Promise<WSRes<T, K>> => {
        let promise: Promise<WSEndpointRes<AnyObj>> = Promise.resolve({});

        if (!this.ws) {
            return promise;
        }

        const body: Partial<WSPacketBody<AnyObj>> = {
            data: data,
        };
        // Add connection ID if handshake is completed
        if (this.id) {
            body.id = this.id;
        }
        // Add token if present
        const authToken = this.getAuthToken();
        const shouldAuthnticate = authToken &&
            ((this.authEndpoints.length == 0) || this.authEndpoints.includes(type as keyof Endpoints));
        if (shouldAuthnticate) {
            body.auth = authToken;
        }
        // Add ack ID if requested
        if (ack) {
            const ackId = this.ackId++;
            body.ackId = ackId;

            promise = new Promise((resolve) => {
                this.pendingAcks.set(ackId, resolve);
            });
        }

        const message = JSON.stringify({ type, body });
        console.log('send: ' + message);
        this.ws.send(message);
        return promise;
    };

    private onOpen = async () => {
        this.internalSend<ReservedEndpoints, '.handshake'>('.handshake', {}, true).then((res) => {
            if (!res.data) {
                return;
            }

            this.id = res.data.id;
            this.isConnected = true;
            if (this.pendingHandshake.resolve) {
                this.pendingHandshake.resolve();
            }
        });

        this._subscription.publish('open', {});
    };

    private onClose = () => {
        this.isConnected = false;

    };

    private onError = (event: Event) => {
        console.log('[SocketClient] error', event);
    };

    private onMessage = (event: MessageEvent) => {
        const msg: {
            type: string,
            body: WSPacketBody<AnyObj>
        } = JSON.parse(event.data);

        console.log('incoming: ', msg);

        if (msg.type == '.ack' && msg.body.ackId) {
            const ack = this.pendingAcks.get(msg.body.ackId);
            if (ack) {
                this.pendingAcks.delete(msg.body.ackId);
                ack({
                    data: msg.body.data,
                    error: msg.body.error,
                });
            }
            return;
        }

        this._subscription.publish(msg.type, {
            error: msg.body.error,
            data: msg.body.data
        });
    };

}